
#include "ppmz2.h"
#include "config.h"
#include "loe.h"
#include <crblib/fileutil.h>
#include <crblib/memutil.h>
#include <crblib/crc32.h>
#include <crblib/tsc.h>

#define PREAMBLE	(1024)

static const ulong PPMZ2_Tag = 0x70707A32; // PPZ2
 
int main(int argc,char * argv[])
{
char *inName,*outName;
ubyte *rawBuf,*compBuf,*decBuf;
int compLen,rawLen;
bool encOnly;
FILE *inFP;
FILE *outFP;
bool encoding;
ulong rawCRC;
char *conditionName = NULL;
FILE *conditionFP = NULL;
ubyte *conditionBuf = NULL;
int conditionLen = 0;
bool doUpdate;

	dprintf("PPMZ2 v0.81 by Charles Bloom, May 2004\n");

	if ( argc < 2 )
	{
		errputs("usage : ppmz2 [options] <in> [out]");
		errputs("options : ");
		errputs("  e  : encode only [also decode and compare] ");
		errputs("  t  : text mode [normal mode]");
		errputs("  l# : set loe type to # [MPS]");
		errputs("  p# : set tune_param to # ");
		errputs("  cF : condition with file F ");
		errputs("  xF : condition with file F and don't update on self ");
		exit(0);
	}

	argv++; argc--;

	inFP = outFP = conditionFP = NULL;
	inName = outName = conditionName = NULL;
	encOnly = false;
	encoding = true;
	doUpdate = true;

	//---------------------------------------
	// process options

	while(argc>0)
	{
	char * str;
		str = *argv++; argc--;
		if ( *str == '-' || *str == '/' )
		{
		str++;
			switch(toupper(*str++))
			{
				case 'L':
					PPMZ2_LOEType = atoi(str);
					if ( PPMZ2_LOEType < LOETYPE_NONE ) PPMZ2_LOEType = LOETYPE_NONE;
					if ( PPMZ2_LOEType >= LOETYPE_COUNT ) PPMZ2_LOEType = LOETYPE_COUNT-1;
					dprintf("got option : loe type = %d\n",PPMZ2_LOEType);
					break;
				case 'E':
					encOnly = true;
					dprintf("got option : only encode, don't decode\n");
					break;
				case 'T':
					PPMZ2_TextMode = true;
					dprintf("got option : text mode\n");
					break;
				case 'P':
					tune_param = atoi(str);
					dprintf("got option : tune_param = %d\n",tune_param);
					break;
				case 'C':
					conditionName = str;
					dprintf("got option : conditionName = %s\n",conditionName);
					break;
				case 'X':
					conditionName = str;
					doUpdate = false;
					dprintf("got option : no update conditionName = %s\n",conditionName);
					break;
				default:
					fprintf(stderr,"unknown option : %c : skipped\n",str[-1]);
					break;
			}
		}
		else
		{
			if ( ! inName ) inName = str;
			else if ( ! outName ) outName = str;
			else
				fprintf(stderr,"extraneous parameter : %s : ignored\n",str);
		}
	}

	//---------------------------------------

	//pushTSC();

	inFP = fopen(inName,"rb");
	if ( ! inFP )
		errexit("couldn't read input file");

	if ( outName )	
	{
		outFP = fopen(outName,"wb");
		if ( ! outFP )
			errexit("couldn't open output file");
	}

	rawLen = flength(inFP);

	encoding = true;

	if ( outFP )
	{
		// see if inFP is compressed
		ulong tag = fgetul(inFP);
		if ( tag == PPMZ2_Tag )
		{
			// it's packed
			rawLen = fgetul(inFP);
			rawCRC = fgetul(inFP);
			encoding = false;
			
			dprintf("unpacking file : %s to %s\n",FilePart(inName),FilePart(outName));
		}
		else
		{
			// it's not
			fseek(inFP,0,SEEK_SET);
			fputul(PPMZ2_Tag,outFP);
			fputul(rawLen,outFP);
			
			dprintf("packing file : %s to %s\n",FilePart(inName),FilePart(outName));
		}
	}
	else
	{
		dprintf("packing file : %s\n",FilePart(inName));
	}

	rawBuf = malloc(rawLen + 1024 + PREAMBLE);
	assert(rawBuf);

	memset(rawBuf,' ',PREAMBLE);
	rawBuf += PREAMBLE;

	if ( encoding )
	{
		FRead(inFP,rawBuf,rawLen);
		fclose(inFP); inFP = NULL;

		rawCRC = crcbuf(rawBuf,rawLen);

		if ( outFP )
			fputul(rawCRC,outFP);
	}

	decBuf = malloc(rawLen + 1024 + PREAMBLE);
	assert(decBuf);

	memset(decBuf,' ',PREAMBLE);
	decBuf += PREAMBLE;

	compBuf = malloc(max(rawLen,conditionLen)*2 + 65536 + PREAMBLE);
	assert(compBuf);

	memset(compBuf,0,PREAMBLE);
	compBuf += PREAMBLE;
	
	if ( conditionName )
	{
		conditionFP = fopen(conditionName,"rb");
		if ( ! conditionFP )
			errexit("couldn't open condition file");
			
		conditionLen = flength(conditionFP);
		
		conditionBuf = malloc(conditionLen + 1024 + PREAMBLE);
		assert(conditionBuf);

		memset(conditionBuf,' ',PREAMBLE);
		conditionBuf += PREAMBLE;

		FRead(conditionFP,conditionBuf,conditionLen);
		fclose(conditionFP); conditionFP = NULL;
	}

	// this time is pretty pointless, but it takes a long time to calibrate the TSC
	//	the first time it's done, so I do it here
	//showPopTSC("setup",stderr);

	if ( encoding )
	{
		if ( conditionBuf )
			compLen = PPMZ2_EncodeArrayWithPrecondition(rawBuf,rawLen,compBuf,conditionBuf,conditionLen,doUpdate);
		else
			compLen = PPMZ2_EncodeArray(rawBuf,rawLen,compBuf);

		dprintf("%-20s : %8d -> %8d = %1.3f bpc\n",FilePart(inName),rawLen,compLen, 
											(compLen*8.0/(double)rawLen));
	}
	else
	{
		// 12 bytes of header
		compLen = flength(inFP) - 12;
		fseek(inFP,12,SEEK_SET);
		FRead(inFP,compBuf,compLen);
		fclose(inFP); inFP = NULL; 
	}

	if ( ! encOnly )
	{
		ulong decCRC;
		bool ok;

		if ( conditionBuf )
			ok = PPMZ2_DecodeArrayWithPrecondition(decBuf,rawLen,compBuf,conditionBuf,conditionLen,doUpdate);
		else
			ok = PPMZ2_DecodeArray(decBuf,rawLen,compBuf);

		if ( ! ok )
			errexit("decode failed");	
	
		decCRC = crcbuf(decBuf,rawLen);
		
		if ( decCRC == rawCRC )
		{
			dprintf("decoded crc matches\n");
		}
		else
		{
			fprintf(stderr,"******* !!!!!! <<<&>>> !!!!!! *******\n");
			dprintf("decoded crc failure : %08X vs %08X\n",rawCRC,decCRC);
		}
	}

	// compare rawBuf and decBuf if we have them both
	if ( encoding && ! encOnly )
	{
		if ( memcmp(decBuf,rawBuf,rawLen) == 0 )
		{
			dprintf("decoded successfully\n");
		}
		else
		{
			fprintf(stderr,"******* !!!!!! <<<&>>> !!!!!! *******\n");
			dprintf("decode failed! %d th bytes differ\n",whichByteDiff(decBuf,rawBuf));
		}
	}

	if ( outFP )
	{
		if ( encoding )
		{
			FWrite(outFP,compBuf,compLen);
		}
		else
		{
			FWrite(outFP,decBuf,rawLen);
		}

		fclose(outFP);
		outFP = NULL;
	}

	//---------------------------------------

	dprintf("done.\n");

	#ifdef _DEBUG
	errputs("press enter");
	fgetc(stdin);
	#endif

return 0;
}
